import logging

logger = logging.getLogger(__name__)

class Led:
    '''
    Module to implement the led operations (only on IPAM400 for now).
    It supports the functions Red, Green, Yellow and Off. Every collor
    could also be set to blink.
    '''
    LEDS_NAME = ('led1', 'led2')
    LEDS_COLOR = ('off', 'red', 'green', 'yellow')

    LED_ON = '255'
    LED_OFF = '0'
    BASE_PATH = '/sys/class/leds/barix:'

    def __init__(self, ledName:str):
        '''
        Instantiate a led object.

        :param str ledName: Led name. Supported values are in LEDS_NAME tuple.

        Raises:
            ValueError: if 'ledName' value is invalid
        '''

        if ledName not in self.LEDS_NAME:
            raise ValueError('Value for ledName not supported')

        self.ledPaths = {
            'red': {
                'brightness': self.BASE_PATH + ledName + ':red/brightness',
                'trigger':    self.BASE_PATH + ledName + ':red/trigger',
                'delay_on':   self.BASE_PATH + ledName + ':red/delay_on',
                'delay_off':  self.BASE_PATH + ledName + ':red/delay_off'
            },
            'green': {
                'brightness': self.BASE_PATH + ledName + ':green/brightness',
                'trigger':    self.BASE_PATH + ledName + ':green/trigger',
                'delay_on':   self.BASE_PATH + ledName + ':green/delay_on',
                'delay_off':  self.BASE_PATH + ledName + ':green/delay_off'
            }
        }

    def __setValue(self, path:str, value:str) -> None:
        try:
            with open(path, 'w') as f:
                f.write(str(value))
        except OSError as e:
            logger.error('Error setting LEDS path {} to value {}: {}'.format(path, value, e))

    def __setBlink(self, ledPath:str, blink:bool, timeOn:int, timeOff:int) -> None:
        trigger = 'none'
        if blink:
            trigger = 'timer'
        self.__setValue(ledPath['trigger'], trigger)
        if blink:
            self.__setValue(ledPath['delay_on'], timeOn)
            self.__setValue(ledPath['delay_off'], timeOff)

    def set(self, color:str, blink:bool=False, blinkTime:tuple=(1000, 1000)) -> None:
        '''
        Change the state of the led.

        :param str color: Led color. Supported values are in LEDS_COLOR tuple.
        :param bool blink: If true, the led will start blinking.
        :param tuple blinkTime: A tuple with the blinking times in milliseconds.
            First position in the tuple is the 'on' time. Second position is for
            the 'off' time.

        Raises:
            ValueError: if 'color' value is invalid
            TypeError: if 'blinkTime' is not a tuple
        '''
        if color not in self.LEDS_COLOR:
            raise ValueError('Leds color not supported')

        if not isinstance(blinkTime, tuple):
            raise TypeError('blinkTime is not a tuple')

        self.__setValue(self.ledPaths['red']['brightness'], self.LED_OFF)
        self.__setValue(self.ledPaths['green']['brightness'], self.LED_OFF)

        if color == 'yellow':
            self.__setBlink(self.ledPaths['red'], blink, blinkTime[0], blinkTime[1])
            self.__setBlink(self.ledPaths['green'], blink, blinkTime[0], blinkTime[1])
            self.__setValue(self.ledPaths['red']['brightness'], self.LED_ON)
            self.__setValue(self.ledPaths['green']['brightness'], self.LED_ON)
        elif color != 'off':
            self.__setBlink(self.ledPaths[color], blink, blinkTime[0], blinkTime[1])
            self.__setValue(self.ledPaths[color]['brightness'], self.LED_ON)

